/* $Id$
 *
 * Primary boot loader
 *
 *  first.S
 *
 *  Copyright (C) 2004  Free Software Initiative of Japan
 *
 * This file is subject to the terms and conditions of the GNU Lesser
 * General Public License version 2.1 (or later).
 *
 */

/*
 * NOTE: Keep this code "position independent", so that this works well
 *       among machines with different memory map.
 */

	.data
	.global start
start:
	.byte	0x7e, 0x1a	/* bl	real_start	*/
	.byte	0xf0, 0x00	/* nop			*/

	.short	0
	.ascii	"LILO"
	.byte	1,0		/* STAGE_FIRST, little endian */

	/* EDIT HERE ! */
	/* Depends LILO's version */
	      /* v--- Major Version    */
	.byte	20, 0
	      /*    ^--- Minor Version */

/* x86 LILO parameters (Not used for other archs ... yet) */
timeout:.short	0		// input timeout
delay:	.short	0		// boot delay
port:	.byte	0		// COM port (0 = unused, 1 = COM1, etc.)
sparam:	.byte	0		// serial port parameters (0 = unused)

/* Timestamp (Filled by LILO command) */
tstamp:	.long	0

/* First descripter sector (Filled by LILO command) */
d1_cx:	.short	0
d1_dx:	.short	0
d1_al:	.byte	0

/* Second descripter sector (Filled by LILO command) */
d2_cx:	.short	0
d2_dx:	.short	0
d2_al:	.byte	0

/* Default command-line sector (Filled by LILO command) */
dc_cx:	.short	0
dc_dx:	.short	0
dc_al:	.byte	0

/* Prompt? (Filled by LILO command) */
prompt:	.byte	0		// indicates whether to always enter prompt
				// (also used as alignment byte)

/* Greeting message length & sector (Filled by LILO command) */
ms_len:	.short	0		// initial greeting message
ms_cx:	.short	0
ms_dx:	.short	0
ms_al:	.byte	0

/* Second descripter sector (Filled by LILO command) */
kt_cx:	.short	0		// keyboard translation table
kt_dx:	.short	0
kt_al:	.byte	0

d_addr:				// second stage sector addresses
	.long	0xffffffff
	.long	0xffffffff
	.long	0xffffffff
	.long	0xffffffff
	.long	0xffffffff
	.long	0xffffffff
	.long	0xffffffff
	.long	0xffffffff
	.long	0xffffffff
	.long	0xffffffff
	.long	0xffffffff
	.long	0xffffffff

	.short	0xffff

	.short	0x0000

	.text
/*
 *	0-3 arguments/results
 *	4-7 call used
 *	8-12 call saved
 *	13 frame pointer
 *	14 link register
 *	15 stack pointer
 *
 *	boot(bios_call_vector, mem_start, mem_size)
 *	
 *	r8   -------------------------- desc
 *	r9   -------------------------- buffer
 *	r10  512
 *	r11  bios_call_vector
 *
 *   on the stack:
 *	base_pointer, mem_size, mem_start
 */
	.align	2
real_start:
	/* This code is written as Position Independent Code. */
	/* All the acess should be indirect. */
	/* Initialize base pointer. */
	mv	r12, r14
	addi	r12, #-4	; address of the label "start"
	;
	mv	r11, r0		; bios_call_vector
	push	r1		; mem_start
	push	r2		; mem_size
	push	r12
	;;
	/* Load second stage loader */
	ldi	r0, #0x4c	; 'L'
	bl.s	put_char
	.align	2
	;; 
	ldi	r8, #52
	add	r8, r12		; descriptor address
	;; 
	mv	r9, r12		; r9 = buffer - 512
	;;
	ldi	r10, #2
	slli	r10, #8		; = 512
	;; 
	bl.s	loop		; jump to loop, move pc --> r14

	.align	2
loop:	mv	r12, r14
	add	r9, r10
	/* A sector address information is five byte long. */
	/* Note that it's not 4-byte aligned. */
	ldub	r1, @r8
	addi	r8, 1
	ldub	r4, @r8
	addi	r8, 1
	ldub	r0, @r8
	addi	r8, 1
	and3	r0, r0, #0x0f
	ldub	r5, @r8
	addi	r8, 1
	slli	r5, #8
	or	r4, r5
	slli	r4, #8
	or	r1, r4
	ldub	r3, @r8
	addi	r8, 1
	;; r0 = dev, r1 = lba
	beqz	r3, done
	;; 
	mv	r2, r9
	mv	r14, r12
	bra	read_sectors	; call to read_sector, back to loop

	.align	2
done:	ldi	r0, #0
	bl.s	flush_cache
	mv	r0, r9
	mv	r1, r11
	pop	r4
	pop	r3
	pop	r2
	push	r4
	add	r4, r10
	jmp	r4

	.align	2
put_char:
	ld	r4, @r11
	jmp	r4

	.align	2
put_string:
	add3	r4, r11, #4
	ld	r5, @r4
	jmp	r5

	.align	2
read_sectors:
	add3	r4, r11, #8
	ld	r5, @r4
	jmp	r5

	.align	2
flush_cache:
	add3	r4, r11, #12
	ld	r5, @r4
	jmp	r5

	.section .rodata
p_table:			/* the partition table */
	.space	16
	.space	16
	.space	16
	.space	16
	.short	0xAA55		/* boot block signature */
	.end
