/* $Id$
 *
 * Secondary boot loader
 *
 *  second.c
 *
 *  Copyright (C) 2004  Free Software Initiative of Japan
 *
 * This file is subject to the terms and conditions of the GNU Lesser
 * General Public License version 2.1 (or later).
 *
 */

/* No variable references */
/* No strings */

#define SD_DESCR1	24
#define SD_DESCR2	29

asm ("bra	start\n\t.fillinsn");

static void inline
put_char (int c, unsigned long bios_call_vector)
{
  register long __r0 __asm__ ("r0") = (long) c;
  register long __r1 __asm__ ("r1") = *(long *)(bios_call_vector);
  register long __r2 __asm__ ("r2");
  register long __r3 __asm__ ("r3");
  register long __r4 __asm__ ("r4");
  register long __r5 __asm__ ("r5");
  register long __r6 __asm__ ("r6");
  register long __r7 __asm__ ("r7");
  register long __r14 __asm__ ("r14");

  asm volatile ("jl	%1\n\t.fillinsn"
		: "=r" (__r0), "=r" (__r1), "=r" (__r2), "=r" (__r3),
		  "=r" (__r4), "=r" (__r5), "=r" (__r6), "=r" (__r7),
		  "=r" (__r14)
		: "0" (__r0), "1" (__r1)
		: "memory");
}

static void inline
put_string (const unsigned char *str, unsigned long bios_call_vector)
{
  register long __r0 __asm__ ("r0") = (long) str;
  register long __r1 __asm__ ("r1") = *(long *)(bios_call_vector + 4);
  register long __r2 __asm__ ("r2");
  register long __r3 __asm__ ("r3");
  register long __r4 __asm__ ("r4");
  register long __r5 __asm__ ("r5");
  register long __r6 __asm__ ("r6");
  register long __r7 __asm__ ("r7");
  register long __r14 __asm__ ("r14");

  asm volatile ("jl	%1\n\t.fillinsn"
		: "=r" (__r0), "=r" (__r1), "=r" (__r2), "=r" (__r3),
		  "=r" (__r4), "=r" (__r5), "=r" (__r6), "=r" (__r7),
		  "=r" (__r14)
		: "0" (__r0), "1" (__r1)
		: "memory");
}

static int inline
read_sectors (int dev, unsigned long lba, unsigned char *buf, int count,
	      unsigned long bios_call_vector)
{
  register long __r0 __asm__ ("r0") = (long) dev;
  register long __r1 __asm__ ("r1") = (long) lba;
  register long __r2 __asm__ ("r2") = (long) buf;
  register long __r3 __asm__ ("r3") = (long) count;
  register long __r4 __asm__ ("r4") = *(long *)(bios_call_vector + 8);
  register long __r5 __asm__ ("r5");
  register long __r6 __asm__ ("r6");
  register long __r7 __asm__ ("r7");
  register long __r14 __asm__ ("r14");

  asm volatile ("jl	%4\n\t.fillinsn"
		: "=r" (__r0), "=r" (__r1), "=r" (__r2), "=r" (__r3),
		  "=r" (__r4), "=r" (__r5), "=r" (__r6), "=r" (__r7),
		  "=r" (__r14)
		: "0" (__r0), "1" (__r1), "2" (__r2), "3" (__r3), "4" (__r4)
		: "memory");

#if 0
  if (__r0 < 0)
    put_string ("ERROR: Sector read\n");
#endif

  return __r0;
}

static void inline
flush_cache (int how, unsigned long bios_call_vector)
{
  register long __r0 __asm__ ("r0") = (long) how;
  register long __r1 __asm__ ("r1") = *(long *)(bios_call_vector+12);
  register long __r2 __asm__ ("r2");
  register long __r3 __asm__ ("r3");
  register long __r4 __asm__ ("r4");
  register long __r5 __asm__ ("r5");
  register long __r6 __asm__ ("r6");
  register long __r7 __asm__ ("r7");
  register long __r14 __asm__ ("r14");

  asm volatile ("jl	%1\n\t.fillinsn"
		: "=r" (__r0), "=r" (__r1), "=r" (__r2), "=r" (__r3),
		  "=r" (__r4), "=r" (__r5), "=r" (__r6), "=r" (__r7),
		  "=r" (__r14)
		: "0" (__r0), "1" (__r1)
		: "memory");
}

static void inline
get_cpu_info (unsigned long *cpu_info, unsigned long bios_call_vector)
{
  register long __r0 __asm__ ("r0") = (long) cpu_info;
  register long __r1 __asm__ ("r1") = *(long *)(bios_call_vector+20);
  register long __r2 __asm__ ("r2");
  register long __r3 __asm__ ("r3");
  register long __r4 __asm__ ("r4");
  register long __r5 __asm__ ("r5");
  register long __r6 __asm__ ("r6");
  register long __r7 __asm__ ("r7");
  register long __r14 __asm__ ("r14");

  asm volatile ("jl	%1\n\t.fillinsn"
		: "=r" (__r0), "=r" (__r1), "=r" (__r2), "=r" (__r3),
		  "=r" (__r4), "=r" (__r5), "=r" (__r6), "=r" (__r7),
		  "=r" (__r14)
		: "0" (__r0), "1" (__r1)
		: "memory");
}

static int inline
get_sector_address (unsigned char *sector_desc, int *devp, unsigned long *lbap)
{
  int count;

  /* Number of sectors */
  count = (int)sector_desc[4];
  if (count == 0)
    return 0;

#if 0
  /* sector_desc[2]: drive number */
  if ((int)sector_desc[2] < 0xc0)
    /* XXX: should return error */
    put_string ("ERROR: Sector address is not in LBA\n");
#endif

  *devp = (int)sector_desc[2] & 0x0f;
  *lbap = sector_desc[0] + (sector_desc[1]<<8) + (sector_desc[3]<<16);

  return count;
}

static int
load_sectors (unsigned char *sector_desc, unsigned char *buf,
	      unsigned long bcv)
{
  int dev;
  unsigned long lba;
  int count;

  count = get_sector_address (sector_desc, &dev, &lba);

  if (count)
    return read_sectors (dev, lba, buf, count, bcv);

  return -1;
}

static unsigned char *
load_sectors_with_maps (unsigned char *desc, int offset,
			unsigned char **buf_p, unsigned char *map_buf,
			unsigned long bcv)
{
  int dev;
  unsigned long lba;
  int i, count;

  if (load_sectors (desc, map_buf, bcv) < 0)
    return 0;

  for (i = offset*5; i<505; i+=5)
    {
      if ((count = get_sector_address (map_buf+i, &dev, &lba)) == 0)
	return 0;

      read_sectors (dev, lba, *buf_p, count, bcv);
      *buf_p += count*512;
    }

  /* There's next map */
  return map_buf+505;
}

#define KERNEL_LOAD_OFFSET	0x00400000 /* 4M-byte */

static void inline
goto_kernel (unsigned long bios_call_vector,
	     unsigned char *main_mem_start,
	     unsigned long mem_size,
	     unsigned char *initrd_start,
	     unsigned long initrd_size)
{
  unsigned char *kernel_params = (unsigned char *)main_mem_start + 0x1000;

  /* XXX
   * We should specify memory size with mem_size
   */

  /* 0xff000400 is "bra +0x1000", because of wrong firmware vector */
  *(unsigned long *)(kernel_params + 0x0000) = 0xff000400; /* mount rd only */
  *(unsigned long *)(kernel_params + 0x0004) = 0x4000;	/* ramdisk flag */
  *(unsigned long *)(kernel_params + 0x0008) = 0x0301;	/* orig root */
  *(unsigned long *)(kernel_params + 0x000c) = 1;	/* loader type */
  *(unsigned long *)(kernel_params + 0x0010) = (unsigned long)initrd_start;
  *(unsigned long *)(kernel_params + 0x0014) = initrd_size;

  get_cpu_info ((unsigned long *)(kernel_params + 0x0018), bios_call_vector);

  kernel_params[0x100] = 'c';
  kernel_params[0x101] = 'o';
  kernel_params[0x102] = 'n';
  kernel_params[0x103] = 's';
  kernel_params[0x104] = 'o';
  kernel_params[0x105] = 'l';
  kernel_params[0x106] = 'e';
  kernel_params[0x107] = '=';
  kernel_params[0x108] = 't';
  kernel_params[0x109] = 't';
  kernel_params[0x10a] = 'y';
  kernel_params[0x10b] = '1';
  kernel_params[0x10c] = ' ';
  kernel_params[0x10d] = 'c';
  kernel_params[0x10e] = 'o';
  kernel_params[0x10f] = 'n';
  kernel_params[0x110] = 's';
  kernel_params[0x111] = 'o';
  kernel_params[0x112] = 'l';
  kernel_params[0x113] = 'e';
  kernel_params[0x114] = '=';
  kernel_params[0x115] = 't';
  kernel_params[0x116] = 't';
  kernel_params[0x117] = 'y';
  kernel_params[0x118] = 'D';
  kernel_params[0x119] = '0';
  kernel_params[0x11a] = ',';
  kernel_params[0x11b] = '1';
  kernel_params[0x11c] = '1';
  kernel_params[0x11d] = '5';
  kernel_params[0x11e] = '2';
  kernel_params[0x11f] = '0';
  kernel_params[0x120] = '0';
  kernel_params[0x121] = 'n';
  kernel_params[0x122] = '8';
  kernel_params[0x123] = 'x';
  kernel_params[0x124] = ' ';
  kernel_params[0x125] = 'r';
  kernel_params[0x126] = 'o';
  kernel_params[0x127] = 'o';
  kernel_params[0x128] = 't';
  kernel_params[0x129] = '=';
  kernel_params[0x12a] = '/';
  kernel_params[0x12b] = 'd';
  kernel_params[0x12c] = 'e';
  kernel_params[0x12d] = 'v';
  kernel_params[0x12e] = '/';
  kernel_params[0x12f] = 'h';
  kernel_params[0x130] = 'd';
  kernel_params[0x131] = 'a';
  kernel_params[0x132] = '1';
  kernel_params[0x133] = ' ';
  kernel_params[0x134] = 'm';
  kernel_params[0x135] = 'e';
  kernel_params[0x136] = 'm';
  kernel_params[0x137] = '=';
  kernel_params[0x138] = '3';
  kernel_params[0x139] = '2';
  kernel_params[0x13a] = 'M';
  kernel_params[0x13b] = '\0';
  kernel_params[0x13c] = '\0';
  kernel_params[0x13d] = '\0';
  kernel_params[0x13e] = '\0';
  kernel_params[0x13f] = '\0';

  {
    register long __r0 __asm__ ("r0") = (long)bios_call_vector;
    register long __r1 __asm__ ("r1") = (long)main_mem_start+KERNEL_LOAD_OFFSET;
    register long __r2 __asm__ ("r2") = (long)mem_size;

    asm volatile ("jmp	r1\n\t.fillinsn"
		  : 
		  : "r" (__r0), "r" (__r1), "r" (__r2)
		  : "memory");
  }
}

void
start (unsigned char *loader_end,
       unsigned long bcv,
       unsigned char *main_mem_start,
       unsigned long mem_size,
       unsigned char *base_pointer)
{
  unsigned char *kernel_image, *initrd_addr, *initrd_start;
  unsigned int initrd_size; 

  put_char ('I', bcv);
  load_sectors (SD_DESCR1+base_pointer, loader_end+0x200, bcv);
  load_sectors (SD_DESCR2+base_pointer, loader_end+0x400, bcv);
  put_char ('L', bcv);
  /* XXX: DEFCMD, KBDTBL */
  /* XXX: checksum */
  put_char ('O', bcv);
  /* XXX: MSG */
  put_char (':', bcv);
  put_char (' ', bcv);
  put_string ((char *)(loader_end+0x200+2), bcv); /* Image name */

  kernel_image = main_mem_start + KERNEL_LOAD_OFFSET;

  {
    unsigned char *desc = loader_end+0x200+2+16+16+4+5; /* kernel image */

    /* Skip two sectors: Fallback command line and options */
    desc = load_sectors_with_maps (desc, 2, &kernel_image, loader_end, bcv);
    put_char ('.', bcv);

    while (desc != 0)
      {
	desc = load_sectors_with_maps (desc, 0, &kernel_image,
				       loader_end, bcv);
	put_char ('.', bcv);
      }
  }
  put_char ('\r', bcv);
  put_char ('\n', bcv);

  initrd_size = *(loader_end+0x200+2+16+16)
    + (*(loader_end+0x200+2+16+16+1) << 8)
    + (*(loader_end+0x200+2+16+16+2) << 16)
    + (*(loader_end+0x200+2+16+16+3) << 24);

  initrd_start = initrd_addr = 0;
  if (initrd_size)
    {
      unsigned char *desc = loader_end+0x200+2+16+16+4; /* initrd image */

      initrd_start
	= initrd_addr
	= (unsigned char *)(((unsigned long)kernel_image
			     + 0x00010fff) & ~0x0fff);
      /* 64k added (.bss for zImage), 4K aligned */

      desc = load_sectors_with_maps (desc, 0, &initrd_addr, loader_end, bcv);
      if (desc != 0)
	{
	  put_char ('i', bcv);
	  put_char ('n', bcv);
	  put_char ('i', bcv);
	  put_char ('t', bcv);
	  put_char ('r', bcv);
	  put_char ('d', bcv);
	  put_char (':', bcv);
	  put_char (' ', bcv);
	  while (desc != 0)
	    {
	      desc = load_sectors_with_maps (desc, 0, &initrd_addr,
					     loader_end, bcv);
	      put_char ('.', bcv);
	    }
	  put_char ('\r', bcv);
	  put_char ('\n', bcv);
	}
    }

  flush_cache (0, bcv);

  goto_kernel (bcv, main_mem_start, mem_size, initrd_start, 
	       initrd_addr - initrd_start);

  /* Never reached */
  for (;;)
    asm volatile ("" : : : "memory");
}
